
#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Jul 14 09:29:36 2022

@author: uzair
"""

from nbody import *
import numpy as np
import gravbody.physics_helper as ph

# np.random.seed(1100)

def orbitCollision(numParticles, orbitDiameter):
    '''Simulates two large orbiting bodies approaching each other'''
    dist = orbitDiameter
    numParticles -= 2
    
    pos = np.zeros((numParticles + 2, 3))
    pos[0] = -dist * np.ones(3)
    pos[1] = dist * np.ones(3)
    velocity = np.zeros((numParticles + 2, 3))
    velocity[0] = np.array([10, 0, 0])
    velocity[0] = np.array([-10, 0, 0])
    masses = np.append(10000000 * np.ones(2), 1 * np.ones(numParticles))
    for i in range(1, numParticles // 2 + 1):
        pos[i + 1] = -dist * np.random.rand(3) - dist / 2
        pos[i + 1] = np.array([pos[i + 1][0], pos[i + 1][1], pos[i + 1][2]])
        normal = (pos[i + 1] - pos[0]) / np.sqrt(np.sum((pos[i + 1] - pos[0]) ** 2))
        velocity[i + 1] = np.sqrt(10000000 / np.sqrt(np.sum((pos[i + 1] - pos[0]) ** 2))) * np.array([-normal[1], normal[0], normal[2]])
    
    for i in range(numParticles // 2 + 1, numParticles + 1):
        pos[i + 1] = dist * np.random.rand(3) + dist / 2
        pos[i + 1] = np.array([pos[i + 1][0], pos[i + 1][1], pos[i + 1][2]])
        normal = (pos[i + 1] - pos[1]) / np.sqrt(np.sum((pos[i + 1] - pos[1]) ** 2))
        velocity[i + 1] = np.sqrt(10000000 / np.sqrt(np.sum((pos[i + 1] - pos[1]) ** 2))) * np.array([-normal[1], normal[0], normal[2]])
    
    nbody = NBody(pos, velocity, masses, 1)
    return nbody
    
def orbit(numParticles, orbitDiameter, bigMass=10000, smallMass=3, **kwargs):
    '''Simulates small particles orbit around a massive particle'''
    
    pos = np.zeros((numParticles, 3))
    velocity = np.zeros((numParticles, 3))
    # Give one particle a much larger mass than all of the other particles
    masses = np.append(bigMass * np.ones(1), smallMass * np.ones(numParticles - 1))
    for i in range(1, numParticles):
        # randomly place particles some distance away from the center
        pos[i] = orbitDiameter * np.random.rand(3) - orbitDiameter / 2
        pos[i, 2] /= 3 # flatten out the orbit 
        
        # get the normal vector between center and particle
        normal = (pos[i] - pos[0]) / ph.dist(pos[i], pos[0])
        perpendicular_dir = np.array([-normal[1], normal[0], normal[2]])
        
        # tangential velocity = sqrt(centripetal acceleration * radius)
        # v = sqrt((large_mass / r^2) * r) = sqrt(large_mass / r) if G is 1
        vel_magnitude = np.sqrt(bigMass / ph.dist(pos[i], pos[0]))
        
        # make particle move fast enough to orbit in a direction tangent to the normal vector
        velocity[i] = vel_magnitude * perpendicular_dir
    
    return NBody(pos, velocity, masses, 1, **kwargs)

def collision(dist=50, masses=np.array([1000.0, 1000])):
    '''Simple collision of two particles with no initial velocity'''
    numParticles = 2
    pos = np.array([[0.0, -dist/2, 0], [0, dist/2, 0]])
    velocity = np.zeros((2, 3))
    
    return NBody(pos, velocity, masses, 1, use_collisions=False, particle_radius=5, barnes_hut=True)

def separation(dist=100, vel=30, masses=np.array([1000, 1000])):
    '''Two particles travelling away at a constant velocity'''
    numParticles = 2
    pos = np.array([[0, -dist/2, 0], [0, dist/2, 0]])
    velocity = np.array([[0, -vel, 0], [0, vel, 0]])
    
    return NBody(pos, velocity, masses, 1)

def single(vel = np.array([[50.0, 0, 0]])):
    pos = np.zeros((1, 3))
    return NBody(pos, vel, np.array([10]), 1)

if __name__ == '__main__':
    sim = orbit(5, 100)
    sim.save(t=0.001, path='animations/force_softening_test', numFrames=100, numFramesPerNotification=100, saveEvery=1)
    



import numpy as np
import gravbody.physics_helper as ph

class BarnesHutNode:
    def __init__(self, center, width):
        '''Barnes-Hut Tree Node Implementation
        
        Center: Position in space that the cube-shaped node is centered on
        Width: Node size
        '''
        self.spatialCenter = center # center of the node (cube-shaped) in space
        self.width = width 
        
        # the center of mass of all particles contained in this node
        self.centerMass = np.zeros(3)
        
        # total mass of particles in this node
        self.totalMass = 0 
        
        self.children = {} 
        self.isLeaf = True # boolean representing if this node has children
        self.velocity = np.zeros(3) # used for collisions; only used when node contains a single particle

    def insert(self, particle_pos, particle_mass, particle_vel):
        '''Inserts a particle into this node based on it's location'''
        
        # if this node doesn't have any particles in it
        if self.isLeaf and self.totalMass == 0:
            # add particle mass to node's mass
            self.updateMass(particle_pos, particle_mass)
            
            # store the particle's velocity (for collisions)
            self.velocity = particle_vel 

        elif not self.isLeaf: # if this node has children
            # add particle mass to node's mass
            self.updateMass(particle_pos, particle_mass)
            
            self.insertInAppropriateChild(particle_pos, particle_mass, particle_vel)

        else: # if this node currently contains only one particle

            # split this node up and assign particles to sub-nodes
            self.insertInAppropriateChild(self.centerMass, self.totalMass, self.velocity)
            self.insertInAppropriateChild(particle_pos, particle_mass, particle_vel)
            
            # add particle mass to node's mass
            self.updateMass(particle_pos, particle_mass)
            
            # this node isn't a leaf anymore, so remove leaf properties
            self.isLeaf = False
            self.velocity = np.zeros(3)
            

    def insertInAppropriateChild(self, particle_pos, particle_mass, particle_vel):
        '''Inserts a node into the appropriate child node based on its position relative to the center of the node'''
        
        # Childname is in the format '[x dir][y dir][z dir]' where each [dir] is a + or - based 
        # on where the particle is relative to the node's center
        #
        # Ex: +++ means that the particle belongs in the child node that has a larger x, y, and z coord
        # relative to the current node's center
        
        diff = particle_pos - self.spatialCenter
        childName =  "+" if (diff[0] > 0) else "-"
        childName += "+" if (diff[1] > 0) else "-"
        childName += "+" if (diff[2] > 0) else "-"
        
        
        # if the node that this particle belongs to is empty (thus not yet created)
        if childName not in self.children.keys():
            # get absolute values from diff vector
            # replace 0s with 1s to avoid divide by 0 errors 
            absDiff = np.abs(diff)
            absDiff[0] = 1 if absDiff[0] == 0 else absDiff[0]
            absDiff[1] = 1 if absDiff[1] == 0 else absDiff[1]
            absDiff[2] = 1 if absDiff[2] == 0 else absDiff[2]
            
            # create a vector storing just the signs (+1, -1) of the diff vector
            signs = diff / absDiff
            
            # if the particle is directly on the center of the node, default to the --- node
            # (arbitrary choice)
            if np.isclose(particle_pos, self.spatialCenter).all():
                signs = -np.ones_like(signs)
                
            # create the node and store it in the children dictionary
            childCenter = signs * self.width / 4 + self.spatialCenter
            self.children[childName] = BarnesHutNode(childCenter, self.width/2)

        # insert the particle in the appropriate child node
        self.children[childName].insert(particle_pos, particle_mass, particle_vel)
    
    def updateMass(self, particle_pos, particle_mass):
        '''Updates the total mass and center of mass of this node after a
        particle has been inserted'''
        
        newTotalMass = self.totalMass + particle_mass
        self.centerMass = (self.centerMass * self.totalMass + particle_pos * particle_mass) / newTotalMass
        self.totalMass = newTotalMass
    
def calcAcceleration(particle, mass, node, threshold, softening=3):
    '''Calculate the net acceleration on a particle based on the Barnes-Hut algorithm
    
    Threshold: value used to determine when a particles are sufficiently far away
    Softening: Value used to reduce the impact of forces when particles get close'''
    accel = np.zeros(3)
    diff = node.centerMass - particle
    distSquared = np.dot(diff, diff) + (softening ** 2)
    if node.isLeaf: # if the node only has one particle
        if not np.isclose(node.centerMass, particle).all():
            accel += node.totalMass * diff / (distSquared ** 1.5)
    else: # if the node contains multiple particles
        if distSquared == 0:
            sd_ratio = threshold + 1
        else:
            sd_ratio = node.width / np.sqrt(distSquared)
            
        if sd_ratio < threshold:
            # if the node is far away, treat all the particles within it as a single mass at its center
            
            accel += node.totalMass * diff / (distSquared ** 1.5)
        else: # if the node is nearby
            # Visit each childnode and determine its effects on this particle
            for child in node.children.values():
                accel += calcAcceleration(particle, mass, child, threshold, softening)
    return accel

def handle_elastic_collisions(particle, vel, mass, node, threshold, radius=5):
    '''Causes particles to bounce off of each other when they get close while
    conserving kinetic energy.
    
    Returns new velocity of the given particle'''

    ### TODO
    ### Double check that this algorithm works
    ### There may be cases where the threshold isn't met even though a collision should happen
    
    dist = ph.dist(node.centerMass, particle)
    if node.isLeaf: # if the node only has one particle
        if not np.isclose(node.centerMass, particle).all():
            if dist < 2 * radius:
                v = ph.elastic_collision(particle, vel, mass, node.centerMass,
                                         node.velocity, node.totalMass)
                return v[0]
    else: # if the node contains multiple particles
        if dist == 0:
            sd_ratio = threshold + 1 # make sure to meet the threshold, since we are very close
        else:
            sd_ratio = node.width / dist
            
        if sd_ratio < threshold:
            # if the node is far away, we can't collide with it
            return vel
        else: # if the node is nearby
            for child in node.children.values():
                # check if we collide with any child nodes
                v = handle_elastic_collisions(particle, vel, mass, child, threshold, radius)
                if (v != vel).any():
                    return v
    
    return vel



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Jul 22 11:45:59 2022

@author: uzair
"""

# Quick Testing Instructions
#
# Run: `python3 main.py S sample_data.txt [output_dir] 100 -t 0.005`
# This creates an orbital simulation
#
# Run: `python3 main.py V [simulation_dir] -e -t 50`
# This visualizes the simulation with an energy plot as well
# Watch the scale of the energy plot: it's zoomed in so it seems like it varies more than it does

import gravbody.nbody as nbody
import gravbody.nbody_view as nbody_view
import argparse

def main():
    parser = argparse.ArgumentParser(description='An N-Body Simulator')

    subparsers = parser.add_subparsers(dest="mode", help="Mode: [S]imulate or [V]iew saved simulation")

    sim = subparsers.add_parser("S")
    view = subparsers.add_parser("V")

    sim.add_argument("input_file", help="file containing initial conditions(pos, vel, mass) of all particles")
    sim.add_argument("output_dir", help="folder to store simulation output")
    sim.add_argument("time", help="the length of time that will be simulated", type=float)
    sim.add_argument("-t", "--time_step", help="the amount of time in between each simulation update", type=float)
    sim.add_argument("-b", "--barnes_hut", 
                    help="Use the barnes-hut algorithm to speed up execution for a large # of particles",
                    action="store_true")
    sim.add_argument("-c", "--collide_radius", help="Allow particles to collide elastically with a specified radius",
                    type=float)
    sim.add_argument("-s", "--save_every", help="Only save 1 frame per n frames generated",
                    type=int, default=1)
    sim.add_argument("-G", "--G_Constant", help="Set value of gravitational constant",
                    type=float, default=1)
    sim.add_argument("-f", "--force_softening", help="Aritifically increase distance to avoid unexpected behaviour near collisions",
                    type=float, default=0)

    view.add_argument("input_dir", help="directory containing simulation data")
    view.add_argument("-s", "--size", help="Set size of display", default=100, type=float)
    view.add_argument("-d", "--use_2d", help="display particles in a 2d projection instead of 3d",
                    action="store_true")
    view.add_argument("-r", "--relative", help="keep view relative to center of mass",
                    action="store_true")
    view.add_argument("-e", "--energy", help="plot the energy of the system",
                    action="store_true")
    view.add_argument("-t", "--time_scale", help="speed up visualization by a given factor",
                    default=1, type=int)
    view.add_argument("-o", "--output_file", help="store the output visualization in a video file")

    args = parser.parse_args()
    try:
        if args.mode == "S":
            t_step = args.time_step if args.time_step is not None else args.time / 1000
            numFrames = int(args.time / t_step / args.save_every)
            
            colliding = args.collide_radius is not None
            
            sim = nbody.NBody.FromFile(args.input_file, barnes_hut=args.barnes_hut,
                                    use_collisions=colliding, particle_radius=args.collide_radius,
                                    G=args.G_Constant, softening=args.force_softening)
            sim.save(t_step, args.output_dir, numFrames, saveEvery=args.save_every)

        elif args.mode == "V":
            view = nbody_view.NBodyView(args.input_dir, args.size, not args.use_2d,
                                        args.relative, args.time_scale, args.energy,
                                        args.output_file)
            view.display()

        else:
            parser.print_help()
    
    except AttributeError:
        parser.print_help()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Jul 22 10:19:30 2022

@author: uzair
"""

import numpy as np

def centerOfMass(masses, pos):
    '''Calculate the center of mass of a group of particles'''
    com = np.sum(pos * masses.reshape(len(masses), 1), axis=0) / np.sum(masses)
    return com

def elastic_collision(p1_pos, p1_vel, p1_mass, p2_pos, p2_vel, p2_mass):
    '''Calculate the new velocities of two colliding particles in an elastic collision
    Assumes that the two given particles are already colliding (doesn't check for collision)
                                                                
    Returns (p1_velocity, p2_velocity)'''
    
    # TODO
    # Figure out where this comes from:
    # https://physics.stackexchange.com/questions/681396/elastic-collision-3d-eqaution
    
    diff = p2_pos - p1_pos
    
    if (diff == 0).all():
        return p1_vel, p2_vel
    
    dist = np.sqrt(np.sum(diff ** 2))
    normal = (p1_pos - p2_pos) / dist
    
    eff_mass = 1 / (1/p1_mass + 1/p2_mass)
    impact_speed = np.dot(normal, (p1_vel - p2_vel))
    impulse_magnitude = 2 * eff_mass * impact_speed
    
    impulse = normal * impulse_magnitude
    return p1_vel - impulse / p1_mass, p2_vel + impulse / p2_mass


def dist(vector1, vector2):
    return np.sqrt(np.sum((vector1 - vector2) ** 2))


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jul 13 10:17:07 2022

@author: uzair
"""

import matplotlib.pyplot as plt
from matplotlib.animation import FuncAnimation, FFMpegWriter
from matplotlib import ticker
from mpl_toolkits.mplot3d import Axes3D
import numpy as np
import sys


class NBodyView:
    def __init__(self, path, size, three_dimensional=True, relativeToCenterOfMass=True,
                 timeScale=1, plot_energy=False, output_file=None):
        
        self.path = path
        self.size = size
        self.three_dimensional = three_dimensional
        self.relativeToCenterOfMass = relativeToCenterOfMass
        self.timeScale = timeScale
        self.plot_energy = plot_energy
        self.output_file = output_file
        self.energyX = np.array([])
        self.energyY = np.array([])
        self.energyAx = None
        
        self.anim = None
        self.end_anim = False
        self.old_com = np.zeros(3)
        self.scatter = None
        self.energy_scatter = None
        self.t = None
        self.masses = None
        
        self.load_anim()

    def update_plot(self, frame_number):
        try:
            data = np.load(self.path + '/' + str(frame_number * self.timeScale) + '.npy')
            pos = data[:, :3]
            vel = data[:, 3:6]
            com = np.sum(pos * self.masses.reshape(len(self.masses), 1), axis=0) / np.sum(self.masses)
            com_vel = (com - self.old_com) / (self.t * self.timeScale)
            self.old_com = com
            
            if self.plot_energy and frame_number != 0:
                energy = self.calc_energy(pos, vel)
                self.energyX = np.append(self.energyX, [frame_number * self.t * self.timeScale])
                self.energyY = np.append(self.energyY, [energy])
                plt.cla()
                self.energyAx.set_title("Energy Plot")
                self.energyAx.set_xlabel("Time")
                self.energyAx.set_ylabel("Energy")
                self.energyAx.scatter(self.energyX, self.energyY, 9, color='blue')
            
            if self.relativeToCenterOfMass:
                pos -= com
            
            if self.three_dimensional:
                self.scatter._offsets3d = pos.T
            else:
                self.scatter.set_offsets(pos[:, :2])
        except FileNotFoundError:
            self.end_anim = True
            # self.anim.event_source.stop()
    
    def frames_gen(self):
        i = 0
        while not self.end_anim:
            yield i
            i += 1
                
    def display(self):
        fig = plt.figure(figsize=(7,7))
        ax = None
        
        if self.plot_energy:
            grid = plt.GridSpec(3, 1, wspace=0.0, hspace=0.5)
            if self.three_dimensional:
                ax = plt.subplot(grid[0:2,0],
                                 xlim=(-self.size, self.size),
                                 ylim=(-self.size, self.size),
                                 zlim=(-self.size, self.size),
                                 projection='3d')
            else:
                ax = plt.subplot(grid[0:2,0],
                                 xlim=(-self.size, self.size),
                                 ylim=(-self.size, self.size))

            self.energyAx = plt.subplot(grid[2:,0])
            self.energyAx.set_title("Total Energy Over Time")
            self.energyAx.set_xlabel("Time")
            self.energyAx.set_ylabel("Energy")
            axis_format = ticker.FormatStrFormatter("%.6f")
            self.energyAx.yaxis.set_major_formatter(axis_format)
        else:
            if self.three_dimensional:
                ax = plt.axes(xlim=(-self.size, self.size),
                              ylim=(-self.size, self.size),
                              zlim=(-self.size, self.size),
                              projection='3d')
            else:
                ax = plt.axes(xlim=(-self.size, self.size),
                              ylim=(-self.size, self.size))
        
        self.scatter = ax.scatter(np.array([]), np.array([]))
        self.anim = FuncAnimation(fig, self.update_plot, frames=self.frames_gen, interval=0.0001, save_count=sys.maxsize)
        
        if self.output_file is None:
            plt.show()
        else:
            w = FFMpegWriter(fps=30)
            self.anim.save(self.output_file, writer=w)
        
    def calc_energy(self, pos, vel):
        KE = 0
        PE = 0
        for i in range(self.masses.shape[0]):
            KE += 0.5 * self.masses[i] * np.sum((vel[i]) ** 2)
            for j in range(self.masses.shape[0]):
                if i != j:
                    PE -= self.masses[i] * self.masses[j] / np.sqrt(np.sum((pos[i] - pos[j]) ** 2))
        # print(KE, PE, end="\t")
        PE /= 2
        
        return KE + PE
        
    def load_anim(self):
        arr = np.load(self.path + '/data.npy')
        self.t = arr[0]
        self.masses = arr[1:]
    
    def isEnergyConserved(self, acceptableError=0.01):
        minEnergy = None
        maxEnergy = None
        
        for i in range(0, 5000):
            data = np.load(self.path + '/' + str(i) + '.npy')
            pos = data[:, :3]
            
            e = self.calc_energy(pos, data[:, 3:6])
            if minEnergy == None or e < minEnergy:
                minEnergy = e
            if maxEnergy == None or e > maxEnergy:
                maxEnergy = e
        
        variability = (maxEnergy - minEnergy) / minEnergy
        
        return variability < acceptableError

if __name__ == '__main__':
    abc = NBodyView('../../animations/3_body_same_mass/0000', 200, plot_energy=True, timeScale=10)
    abc.display()
    
    # for i in range(1075, 1076):
    #     abc = NBodyView('../../animations/3_body_same_mass/%04d' % i, 40)
    #     if i % 50 == 0:
    #         print('-')
    #     if (not abc.isEnergyConserved()):
    #         print(i)
            



import numpy as np
import gravbody.barnes_hut as bh
import gravbody.physics_helper as ph
import time

class NBody:
    def __init__(self, pos, vel, mass, G, barnes_hut=True, use_collisions=False,
                 softening=0, particle_radius=None):
        '''Implementation of an n-body system
        
        pos and vel are stored as an N x 3 matrix
        mass is a 1-dimensional numpy array'''
        
        self.numParticles = pos.shape[0]
        self.pos = pos
        self.velocity = vel
        self.mass = mass
        self.G = G
        self.barnes_hut = barnes_hut
        self.use_collisions = use_collisions
        self.softening = softening
        self.particle_radius = particle_radius
        
        self.oldAccel = np.zeros(3)
        self.justCollided = [False for i in range(self.numParticles)]
    
    @classmethod
    def FromFile(cls, path, *args, **kwargs):
        '''Reads in a space separated file of particle data
        Format should be: [posX] [posY] [posZ] [velX] [velY] [velZ] [mass] on every line'''
        
        data = np.loadtxt(path)
        pos = data[:, :3]
        vel = data[:, 3:6]
        mass = data[:, 6].reshape(data.shape[0])
        
        return cls(pos, vel, mass, *args, **kwargs)
    
    def leapfrogKickDrift(self, t):
        self.velocity += self.oldAccel * (t/2)
        self.pos += self.velocity * t
        
    def leapfrogFinalKick(self, newAccel, t):
        self.velocity += newAccel * (t / 2)
        self.oldAccel = newAccel
    
    def naive_step(self, t):
        '''Basic O(n^2) method to update particle motion'''
        
        self.leapfrogKickDrift(t)
        
        accel = np.zeros((self.numParticles, 3))
        newVelocities = np.array(self.velocity)
        for i in range(self.numParticles):
            for j in range(self.numParticles):
                if i != j: 
                    # find the difference between the two particles' positions
                    diff = self.pos[j] - self.pos[i]
    
                    # calculate the distance based on the vector between them
                    distSquared = np.sum(diff ** 2) + self.softening**2

                    accel[i] += self.G * self.mass[j] * diff / (distSquared ** 1.5)
        
        self.leapfrogFinalKick(accel, t)
        
        
        # Handle Collisions
        if self.use_collisions: 
            experiencedCollision = [False for i in range(self.numParticles)]
            newVelocities = np.array(self.velocity)
            for i in range(self.numParticles):
                for j in range(self.numParticles):
                    if i != j:
                        dist = ph.dist(self.pos[i], self.pos[j])
                        collided = dist < 2 * self.particle_radius
                        if collided and not self.justCollided[i]:
                            newVel = ph.elastic_collision(self.pos[i], self.velocity[i], self.mass[i],
                                                          self.pos[j], self.velocity[j], self.mass[j])[0]
                            newVelocities[i] = newVel
                        
                        if collided:
                            experiencedCollision[i] = True
                            break
                        
            self.justCollided = experiencedCollision
            self.velocity = newVelocities
    
    def naive_vectorized_step(self, t):
        
        ### INSPIRED BY https://medium.com/swlh/create-your-own-n-body-simulation-with-python-f417234885e9
        
        self.leapfrogKickDrift(t)
        
        x = self.pos[:,0:1]
        y = self.pos[:,1:2]
        z = self.pos[:,2:3]
        
        # matrix that stores all pairwise particle separations: r_j - r_i
        dx = x.T - x
        dy = y.T - y
        dz = z.T - z
        
        # matrix that stores 1/r^3 for all particle pairwise particle separations 
        distSquared = dx**2 + dy**2 + dz**2 + self.softening**2
        inv_r3 = np.power(distSquared, -1.5, out=np.zeros_like(distSquared), where=distSquared!=0)
        ax = self.G * (dx * inv_r3) @ self.mass
        ay = self.G * (dy * inv_r3) @ self.mass
        az = self.G * (dz * inv_r3) @ self.mass
        
        # pack together the acceleration components
        accel = np.stack((ax, ay, az), axis=1)
        self.leapfrogFinalKick(accel, t)
        
        if self.use_collisions: 
            experiencedCollision = [False for i in range(self.numParticles)]
            newVelocities = np.array(self.velocity)
            for i in range(self.numParticles):
                for j in range(self.numParticles):
                    if i != j:
                        dist = ph.dist(self.pos[i], self.pos[j])
                        collided = dist < 2 * self.particle_radius
                        if collided and not self.justCollided[i]:
                            newVel = ph.elastic_collision(self.pos[i], self.velocity[i], self.mass[i],
                                                          self.pos[j], self.velocity[j], self.mass[j])[0]
                            newVelocities[i] = newVel
                        
                        if collided:
                            experiencedCollision[i] = True
                            break
                        
            self.justCollided = experiencedCollision
            self.velocity = newVelocities
    
    def barnes_hut_step(self, t):
        '''Barnes-Hut Algorithm Implementation'''
        
        self.leapfrogKickDrift(t)
        
        # set tree size based on the maximum dist from center of mass to any particle
        com = ph.centerOfMass(self.mass, self.pos)
        maxDist = np.sqrt(np.max(np.sum((self.pos - com) ** 2, 1))) + 1
            
        # Create the tree structure
        root = bh.BarnesHutNode(com, maxDist)
        for i in range(self.numParticles):
            root.insert(self.pos[i], self.mass[i], self.velocity[i])
            
        # Calculate accelerations for each particle
        accel = np.zeros((self.numParticles, 3))
        for i in range(self.numParticles):
            accel[i] = self.G * bh.calcAcceleration(self.pos[i], self.mass[i], root, 1, self.softening)
        self.leapfrogFinalKick(accel, t)
        
        # Handle collisions
        if self.use_collisions:
            experiencedCollision = [False for i in range(self.numParticles)]
            newVelocities = np.array(self.velocity)
            for i in range(self.numParticles):
                newVel = bh.handle_elastic_collisions(self.pos[i], self.velocity[i], self.mass[i], root, 0.5, self.particle_radius)
                collided = (newVel != self.velocity[i]).any()
                if collided and not self.justCollided[i]:
                    newVelocities[i] = newVel
                    
                if collided:
                    experiencedCollision[i] = True
                
            self.justCollided = experiencedCollision
            self.velocity = newVelocities
        
    def advanceSimulation(self, t):
        if self.barnes_hut:
            self.barnes_hut_step(t)
        else:
            self.naive_vectorized_step(t)
    
    def save(self, t, path, numFrames, numFramesPerNotification=None, saveEvery=1):
        '''Saves data from nbody model into animation frame files to be played back later'''
        
        # store the timestep and mass in a data file since they are the same for
        # each frame
        with open(path + '/' + 'data.npy', 'wb') as f:
            data = np.concatenate((np.array([t * saveEvery]), self.mass))
            np.save(f, data)
        
        t1 = time.time()
        start = t1
        for i in range(numFrames):
            with open(path + '/' + str(i) + '.npy', 'wb') as f:
                # combine position and velocity into a single array and save it
                data = np.concatenate((self.pos, self.velocity), axis=1)
                np.save(f, data)
            
            # skip over frames that we aren't saving
            for j in range(saveEvery):
                self.advanceSimulation(t)
            
            # print an update every few frames
            if numFramesPerNotification is not None:
                if (i + 1) % numFramesPerNotification == 0:
                    t2 = time.time()
                    print("Completed", i + 1, "frames!        Time per frame: %.2f s" %
                          ((t2 - t1) / numFramesPerNotification))
                    t1 = time.time()
                
        end = time.time()        
        print("Simulation complete!")
        print("Generated", numFrames, "frames in %.2f seconds" % (end - start))
    
    def calc_energy(self):
        '''Calculates the total energy of the current state of the n body system'''
        
        KE = 0
        PE = 0
        for i in range(self.numParticles):
            KE += 0.5 * self.mass[i] * np.sum((self.velocity[i]) ** 2)
            for j in range(self.numParticles):
                if i != j:
                    PE -= self.G * self.mass[i] * self.mass[j] / np.sqrt(np.sum((self.pos[i] - self.pos[j]) ** 2))
        # print(KE, PE, end="\t")
        PE /= 2
        
        return KE + PE


from setuptools import setup, find_packages

# uses a markdown file README, if you don't have that then 
# delete the .md or something.
with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="gravbody",
    version="0.1.0",
    author="Uzair Nawaz",
    author_email="uzairn@icloud.com",
    description="A gravitational n-body simulation package with features such as a barnes-hut tree, visualization tools, and CLI.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://codeberg.org/uzairn/nbody",
    #project_urls={
    #    "Bug Tracker": "https://github.com/pypa/sampleproject/issues",
    #},

    # license description and other stuff. 
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Operating System :: OS Independent",
    ],

    # required packages. I wouldn't list packages that come default to 
    # python standard library such as argparse. If they are managed via
    # pip then list them
    install_requires=[
        'numpy',
        'matplotlib'
    ],

    # command line hook setup also the way it is working 
    # here is the sourcefile "mainfile" (yours is currently main.py)
    # has a method named main(). It may work that you can just
    # delete the ":main" portion and your code runs smoothly. If not then
    # you will have to put your mainfile (excluding imports) inside of a main method 
    # -- you could then further modularize the creation of the argparser to a submethod
    # which may be nice. It might be necessary to add an if __name__ == "__main__" to the
    # bottom of the file as well -- I have never tested a different approach
    entry_points={
        'console_scripts': [
            'gravbody = gravbody.main:main'
        ]
    },

    packages=find_packages(),

    # this line may be useless if you don't have
    # package data inside src/nbody 
    # maybe documentation should go in there idk? It will only affect 
    # people who install your package via pip and not from source and 
    # don't reference the publically available source code which is linked above
    include_package_data = True,

    # may not be true for you but can't hurt
    python_requires=">=3.6",
)
